@@ -1,5 +1,5 @@
 /* pinentry.c - The PIN entry support library
-   Copyright (C) 2002 g10 Code GmbH
+   Copyright (C) 2002, 2003 g10 Code GmbH
    
    This file is part of PINENTRY.
    
@@ -38,6 +38,10 @@
 #include "pinentry.h"
 
 
+/* Keep the name of our program here. */
+static char this_pgmname[50]; 
+
+
 struct pinentry pinentry =
   {
     NULL,	/* Description.  */
@@ -55,15 +59,15 @@ struct pinentry pinentry =
     0,		/* Debug mode.  */
     0,		/* Enhanced mode.  */
     1,		/* Global grab.  */
-    0		/* Result.  */
+    0,		/* Parent Window ID.  */
+    0,		/* Result.  */
+    0           /* Locale error flag. */
   };
 
 
 char *
 pinentry_utf8_to_local (char *lc_ctype, char *text)
 {
-  char *old_ctype;
-  char *target_encoding;
   iconv_t cd;
   char *input = text;
   size_t input_len = strlen (text) + 1;
@@ -71,17 +75,25 @@ pinentry_utf8_to_local (char *lc_ctype, char *text)
   size_t output_len;
   char *output_buf;
   size_t processed;
+  char *old_ctype;
+  char *target_encoding;
 
   /* If no locale setting could be determined, simply copy the
      string.  */
   if (!lc_ctype)
-    return strdup (text);
+    {
+      fprintf (stderr, "%s: no LC_CTYPE known - assuming UTF-8\n",
+               this_pgmname);
+      return strdup (text);
+    }
 
   old_ctype = strdup (setlocale (LC_CTYPE, NULL));
   if (!old_ctype)
     return NULL;
   setlocale (LC_CTYPE, lc_ctype);
   target_encoding = nl_langinfo (CODESET);
+  if (!target_encoding)
+    target_encoding = "?";
   setlocale (LC_CTYPE, old_ctype);
   free (old_ctype);
 
@@ -94,20 +106,24 @@ pinentry_utf8_to_local (char *lc_ctype, char *text)
   cd = iconv_open (target_encoding, "UTF-8");
   if (cd == (iconv_t) -1)
     {
-      free (output);
+      fprintf (stderr, "%s: can't convert from UTF-8 to %s: %s\n",
+               this_pgmname, target_encoding, strerror (errno));
+      free (output_buf);
       return NULL;
     }
   processed = iconv (cd, &input, &input_len, &output, &output_len);
   iconv_close (cd);
   if (processed == (size_t) -1 || input_len)
     {
+      fprintf (stderr, "%s: error converting from UTF-8 to %s: %s\n",
+               this_pgmname, target_encoding, strerror (errno));
       free (output_buf);
       return NULL;
     }
   return output_buf;
 }
 
-/* Convert TEXT whcih is encoded according to LC_CTYPE to UTF-8.  With
+/* Convert TEXT which is encoded according to LC_CTYPE to UTF-8.  With
    SECURE set to true, use secure memory for the returned buffer.
    Return NULL on error. */
 char *
@@ -127,6 +143,8 @@ pinentry_local_to_utf8 (char *lc_ctype, char *text, int secure)
      string.  */
   if (!lc_ctype)
     {
+      fprintf (stderr, "%s: no LC_CTYPE known - assuming UTF-8\n",
+               this_pgmname);
       output_buf = secure? secmem_malloc (input_len) : malloc (input_len);
       if (output_buf)
         strcpy (output_buf, input);
@@ -150,18 +168,24 @@ pinentry_local_to_utf8 (char *lc_ctype, char *text, int secure)
   cd = iconv_open ("UTF-8", source_encoding);
   if (cd == (iconv_t) -1)
     {
+      fprintf (stderr, "%s: can't convert from %s to UTF-8: %s\n",
+               this_pgmname, source_encoding? source_encoding : "?",
+               strerror (errno));
       if (secure)
-        secmem_free (output);
+        secmem_free (output_buf);
       else
-        free (output);
+        free (output_buf);
       return NULL;
     }
   processed = iconv (cd, &input, &input_len, &output, &output_len);
   iconv_close (cd);
   if (processed == (size_t) -1 || input_len)
     {
+      fprintf (stderr, "%s: error converting from %s to UTF-8: %s\n",
+               this_pgmname, source_encoding? source_encoding : "?",
+               strerror (errno));
       if (secure)
-        secmem_free (output);
+        secmem_free (output_buf);
       else
         free (output_buf);
       return NULL;
@@ -194,21 +218,32 @@ pinentry_setbufferlen (pinentry_t pin, int len)
 }
 
 
-/* Initialize the secure memory subsystem, drop privileges and
-   return.  Must be called early.  */
+/* Initialize the secure memory subsystem, drop privileges and return.
+   Must be called early. */
 void
-pinentry_init (void)
+pinentry_init (const char *pgmname)
 {
+  /* Store away our name. */
+  if (strlen (pgmname) > sizeof this_pgmname - 2)
+    abort ();
+  strcpy (this_pgmname, pgmname);
+
   /* Initialize secure memory.  1 is too small, so the default size
      will be used.  */
   secmem_init (1);
   secmem_set_flags (SECMEM_WARN);
   drop_privs ();
+
+  if (atexit (secmem_term))
+    /* FIXME: Could not register at-exit function, bail out.  */
+    ;
+
+  assuan_set_malloc_hooks (secmem_malloc, secmem_realloc, secmem_free);
 }
 
 /* Simple test to check whether DISPLAY is set or the option --display
    was given.  Used to decide whether the GUI or curses should be
-   initialized. */
+   initialized.  */
 int
 pinentry_have_display (int argc, char **argv)
 {
@@ -225,20 +260,20 @@ pinentry_have_display (int argc, char **argv)
 static void 
 usage (void)
 {
-  /* FIXME: replace the "?" by the real program name. */
-  fprintf (stderr, "Usage: %s [OPTION]...\n\
-Ask securely for a secret and print it to stdout.\n\
-\n\
-      --display DISPLAY Set the X display\n\
-      --ttyname PATH    Set the tty terminal node name\n\
-      --ttytype NAME    Set the tty terminal type\n\
-      --lc-ctype        Set the tty LC_CTYPE value\n\
-      --lc-messages     Set the tty LC_MESSAGES value\n\
-  -e, --enhanced        Ask for timeout and insurance, too\n\
-  -g, --no-global-grab  Grab keyboard only while window is focused\n\
-  -d, --debug           Turn on debugging output\n\
-      --help            Display this help and exit\n\
-      --version         Output version information and exit\n", "?");
+  fprintf (stderr, "Usage: %s [OPTION]...\n"
+"Ask securely for a secret and print it to stdout.\n"
+"\n"
+"      --display DISPLAY Set the X display\n"
+"      --ttyname PATH    Set the tty terminal node name\n"
+"      --ttytype NAME    Set the tty terminal type\n"
+"      --lc-ctype        Set the tty LC_CTYPE value\n"
+"      --lc-messages     Set the tty LC_MESSAGES value\n"
+"  -e, --enhanced        Ask for timeout and insurance, too\n"
+"  -g, --no-global-grab  Grab keyboard only while window is focused\n"
+"      --parent-wid	 Parent window ID (for positioning)\n"
+"  -d, --debug           Turn on debugging output\n"
+"  -h, --help            Display this help and exit\n"
+"      --version         Output version information and exit\n", this_pgmname);
 }
 
 
@@ -252,40 +287,54 @@ pinentry_parse_opts (int argc, char *argv[])
   int opt_help = 0;
   int opt_version = 0;
   struct option opts[] =
-    {{ "debug", no_argument, &pinentry.debug, 1 },
-     { "display", required_argument, 0, 'D' },
-     { "ttyname", required_argument, 0, 'T' },
-     { "ttytype", required_argument, 0, 'N' },
-     { "lc-ctype", required_argument, 0, 'C' },
+    {{ "debug", no_argument,             0, 'd' },
+     { "display", required_argument,     0, 'D' },
+     { "ttyname", required_argument,     0, 'T' },
+     { "ttytype", required_argument,     0, 'N' },
+     { "lc-ctype", required_argument,    0, 'C' },
      { "lc-messages", required_argument, 0, 'M' },
-     { "enhanced", no_argument, &pinentry.enhanced, 1 },
-     { "no-global-grab", no_argument, &pinentry.grab, 0 },
-     { "help", no_argument, &opt_help, 1 },
+     { "enhanced", no_argument,          0, 'e' },
+     { "no-global-grab", no_argument,    0, 'g' },
+     { "parent-wid", required_argument,  0, 'W' },
+     { "help", no_argument,              0, 'h' },
      { "version", no_argument, &opt_version, 1 },
      { NULL, 0, NULL, 0 }};
   
-  while ((opt = getopt_long (argc, argv, "deg", opts, NULL)) != -1)
+  while ((opt = getopt_long (argc, argv, "degh", opts, NULL)) != -1)
     {
       switch (opt)
         {
         case 0:
         case '?':
           break;
+        case 'd':
+          pinentry.debug = 1;
+          break;
+        case 'e':
+          pinentry.enhanced = 1;
+          break;
+        case 'g':
+          pinentry.grab = 0;
+          break;
+        case 'h':
+          opt_help = 1;
+          break;
+
 	case 'D':
+          /* Note, this is currently not used because the GUI engine
+             has already been initialized when parsing these options. */
 	  pinentry.display = strdup (optarg);
 	  if (!pinentry.display)
 	    {
-	      /* XXX Program name.  */
-	      fprintf (stderr, "pinentry: %s\n", strerror (errno));
+	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
 	      exit (EXIT_FAILURE);
 	    }
-	  break;
+	  break; 
 	case 'T':
 	  pinentry.ttyname = strdup (optarg);
 	  if (!pinentry.ttyname)
 	    {
-	      /* XXX Program name.  */
-	      fprintf (stderr, "pinentry: %s\n", strerror (errno));
+	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -293,8 +342,7 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.ttytype = strdup (optarg);
 	  if (!pinentry.ttytype)
 	    {
-	      /* XXX Program name.  */
-	      fprintf (stderr, "pinentry: %s\n", strerror (errno));
+	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -302,8 +350,7 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.lc_ctype = strdup (optarg);
 	  if (!pinentry.lc_ctype)
 	    {
-	      /* XXX Program name.  */
-	      fprintf (stderr, "pinentry: %s\n", strerror (errno));
+	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -311,13 +358,18 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.lc_messages = strdup (optarg);
 	  if (!pinentry.lc_messages)
 	    {
-	      /* XXX Program name.  */
-	      fprintf (stderr, "pinentry: %s\n", strerror (errno));
+	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
+	case 'W':
+	  pinentry.parent_wid = atoi (optarg);
+	  /* FIXME: Add some error handling.  Use strtol.  */
+	  break;
+
         default:
-          /* XXX Should never happen.  */
+          fprintf (stderr, "%s: oops: option not handled\n", this_pgmname);
+	  break;
         }
     }
   if (opt_version) 
@@ -340,11 +392,10 @@ option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
     pinentry.grab = 1;
   else if (!strcmp (key, "debug-wait"))
     {
-      /* XXX Program name.  */
-      fprintf (stderr, "pinentry: waiting for debugger - my pid is %u ...\n",
-	       (unsigned int) getpid());
+      fprintf (stderr, "%s: waiting for debugger - my pid is %u ...\n",
+	       this_pgmname, (unsigned int) getpid());
       sleep (*value?atoi (value):5);
-      fprintf (stderr, "pinentry: ... okay\n");
+      fprintf (stderr, "%s: ... okay\n", this_pgmname);
     }
   else if (!strcmp (key, "display"))
     {
@@ -386,6 +437,11 @@ option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
       if (!pinentry.lc_messages)
 	return ASSUAN_Out_Of_Core;
     }
+  else if (!strcmp (key, "parent-wid"))
+    {
+      pinentry.parent_wid = atoi (value);
+      /* FIXME: Use strtol and add some error handling.  */
+    }
   else
     return ASSUAN_Invalid_Option;
   return 0;
@@ -438,7 +494,7 @@ cmd_setprompt (ASSUAN_CONTEXT ctx, char *line)
   if (!newp)
     return ASSUAN_Out_Of_Core;
 
-  strcpy (newp, line);
+  strcpy_escaped (newp, line);
   if (pinentry.prompt)
     free (pinentry.prompt);
   pinentry.prompt = newp;
@@ -472,7 +528,7 @@ cmd_setok (ASSUAN_CONTEXT ctx, char *line)
   if (!newo)
     return ASSUAN_Out_Of_Core;
 
-  strcpy (newo, line);
+  strcpy_escaped (newo, line);
   if (pinentry.ok)
     free (pinentry.ok);
   pinentry.ok = newo;
@@ -489,7 +545,7 @@ cmd_setcancel (ASSUAN_CONTEXT ctx, char *line)
   if (!newc)
     return ASSUAN_Out_Of_Core;
 
-  strcpy (newc, line);
+  strcpy_escaped (newc, line);
   if (pinentry.cancel)
     free (pinentry.cancel);
   pinentry.cancel = newc;
@@ -511,6 +567,7 @@ cmd_getpin (ASSUAN_CONTEXT ctx, char *line)
       pinentry.prompt = "PIN:";
       set_prompt = 1;
     }
+  pinentry.locale_err = 0;
 
   result = (*pinentry_cmd_handler) (&pinentry);
   if (pinentry.error)
@@ -522,30 +579,28 @@ cmd_getpin (ASSUAN_CONTEXT ctx, char *line)
     pinentry.prompt = NULL;
 
   if (result < 0)
-    return ASSUAN_Canceled;
-
-  if (result && pinentry.pin)
     {
-      char *p;
-
-      p = pinentry_local_to_utf8 (pinentry.lc_ctype, pinentry.pin, 1);
-      if (p)
-        {
-          result = assuan_send_data (ctx, p, strlen (p));
-          secmem_free (p);
-        }
-      else /* Most likely we can't convert between the character sets. */
-        result = ASSUAN_Invalid_Data;
+      if (pinentry.pin)
+	{
+	  secmem_free (pinentry.pin);
+	  pinentry.pin = NULL;
+	}
+      return pinentry.locale_err? ASSUAN_Locale_Problem: ASSUAN_Canceled;
     }
-  else
-    result = assuan_send_data (ctx, pinentry.pin, result);
+
+  if (result)
+    {
+      result = assuan_send_data (ctx, pinentry.pin, result);
+      if (!result)
+	result = assuan_send_data (ctx, NULL, 0);
+    }
+
   if (pinentry.pin)
     {
       secmem_free (pinentry.pin);
       pinentry.pin = NULL;
     }
-  if (!result)
-    result = assuan_send_data (ctx, NULL, 0);
+
   return result;
 }
 
@@ -555,6 +610,7 @@ cmd_confirm (ASSUAN_CONTEXT ctx, char *line)
 {
   int result;
 
+  pinentry.locale_err = 0;
   result = (*pinentry_cmd_handler) (&pinentry);
   if (pinentry.error)
     {
@@ -562,7 +618,9 @@ cmd_confirm (ASSUAN_CONTEXT ctx, char *line)
       pinentry.error = NULL;
     }
 
-  return result ? 0 : ASSUAN_Not_Confirmed;
+  return result ? 0
+                : (pinentry.locale_err? ASSUAN_Locale_Problem
+                                      : ASSUAN_Not_Confirmed);
 }
 
 
@@ -611,6 +669,10 @@ pinentry_loop (void)
   int filedes[2];
   ASSUAN_CONTEXT ctx;
 
+  /* Extra check to make sure we have dropped privs. */
+  if (getuid() != geteuid())
+    abort ();
+
   /* For now we use a simple pipe based server so that we can work
      from scripts.  We will later add options to run as a daemon and
      wait for requests on a Unix domain socket.  */
@@ -619,16 +681,15 @@ pinentry_loop (void)
   rc = assuan_init_pipe_server (&ctx, filedes);
   if (rc)
     {
-      /* XXX Program name.  */
-      fprintf (stderr, "pinentry: failed to initialize the server: %s\n",
-               assuan_strerror(rc));
+      fprintf (stderr, "%s: failed to initialize the server: %s\n",
+               this_pgmname, assuan_strerror(rc));
       return -1;
     }
   rc = register_commands (ctx);
   if (rc)
     {
-      fprintf (stderr, "pinentry: failed to the register commands with Assuan: %s\n",
-               assuan_strerror(rc));
+      fprintf (stderr, "%s: failed to the register commands with Assuan: %s\n",
+               this_pgmname, assuan_strerror(rc));
       return -1;
     }
 
@@ -644,17 +705,16 @@ pinentry_loop (void)
           break;
       else if (rc)
         {
-	  /* XXX Program name.  */
-          fprintf (stderr, "pinentry: Assuan accept problem: %s\n",
-                   assuan_strerror (rc));
+          fprintf (stderr, "%s: Assuan accept problem: %s\n",
+                   this_pgmname, assuan_strerror (rc));
           break;
         }
       
       rc = assuan_process (ctx);
       if (rc)
         {
-          fprintf (stderr, "pinentry: Assuan processing failed: %s\n",
-                   assuan_strerror (rc));
+          fprintf (stderr, "%s: Assuan processing failed: %s\n",
+                   this_pgmname, assuan_strerror (rc));
           continue;
         }
     }
